"""
Wrester for obtaining metadata for a given spl observation for UI display.
"""
import json

# pylint: disable=E0401, E0402, R0903
import logging

import pendulum

from aat_wrest.utilities import MDDBConnector, TIME_ZONE, format_interval


class WrestObservationMetadata:
    """
    Class for extracting observation metadata
    """

    def __init__(self, connection: MDDBConnector, sdm_id: str):
        self.logger = logging.getLogger("aat_wrest")
        self.conn = connection
        self.sdm_id = sdm_id

    def wrest_observation_info(self) -> json:
        """
        Given a sdm_id, query the Metadata DB for observation data related to an ingestion-complete event

        :param sdm_id:
        :return science_product_locator:
        """
        metadata = {}
        sql = f"""
        SELECT science_product_locator,
         band_code,
         configuration,
         starttime,
         endtime
         FROM execution_blocks
         WHERE ngas_fileset_id = %(sdm_id)s
        """
        try:
            cursor = self.conn.cursor()
            cursor.execute(sql, {"sdm_id": self.sdm_id})
            data = cursor.fetchall()
            if data:
                metadata = json.dumps(
                    {
                        "spl": data[0][0],
                        "bands": data[0][1],
                        "array_config": data[0][2],
                        "obs_start_time": data[0][3],
                        "obs_end_time": data[0][4],
                    }
                )
            else:
                self.logger.error(
                    f"ERROR: aat-wrest query returned no results!"
                    f" The database appears to be missing information for SDM id {self.sdm_id}"
                )
        finally:
            self.conn.close()
        return metadata

    #
    # currently unused. Was intended for analyst list display of waiting capability requests/executions
    #
    def wrest_observation_time_info(self) -> json:
        """
        Given a product locator, reports the project code, observation length,
        and time in queue for an incomplete observation.

        :return:
        """

        sql = """
        SELECT project_code, mjd_to_timestamp(starttime), mjd_to_timestamp(endtime)
        FROM execution_blocks
        WHERE science_product_locator = %(spl)s
        """
        make_json = {}
        try:
            cursor = self.conn.cursor()
            cursor.execute(sql, {"spl": self.spl})
            data = cursor.fetchall()

            if data:
                current_time = pendulum.now().in_timezone(TIME_ZONE)
                start_time = data[0][1]
                end_time_raw = data[0][2]

                pend_endtime = pendulum.instance(end_time_raw)
                print(pend_endtime)
                print(current_time)

                since = current_time.diff(pend_endtime).as_interval()
                print(since)

                make_json = json.dumps(
                    {
                        "projectCode": data[0][0],
                        "timeObserved": start_time,
                        "timeInQueue": format_interval(since),
                    },
                    default=str,
                )
            else:
                self.logger.error(
                    f"ERROR: Query returned no results!"
                    f" The database appears to be missing information for SPL {self.spl}!"
                )
        finally:
            self.conn.close()

        return make_json
