"""
Allows for manual intervention into running jobs.
Current primary use case: Several jobs error out before releasing their capability engine,
thereby locking the entire workspaces system against processing further executions.
"""
import argparse
import logging
import sys

# pylint: disable=E0401, W1203

from sqlalchemy.orm import Session, sessionmaker
from sqlalchemy import create_engine
from pycapo import CapoConfig

from messaging.messenger import MessageSender
from workspaces.capability.services.capability_info import CapabilityInfo
from workspaces.capability.services.interfaces import CapabilityInfoIF
from workspaces.workflow.services.interfaces import WorkflowInfoIF
from workspaces.workflow.services.workflow_info import WorkflowInfo

logger = logging.Logger("system_mediator")
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler(sys.stdout))


class Arbiter:
    """
    Class that intercedes in running executions
    """

    def __init__(self, info: WorkflowInfoIF, cinfo: CapabilityInfoIF, nsp: argparse.Namespace):
        self.info = info
        self.cinfo = cinfo
        self.message_router = MessageSender("workflow")
        self.which_option(nsp)

    def which_option(self, nsp: argparse.Namespace):
        """
        Determine which option was chosen by the user and set the required parameters
        :param nsp: argparse Namespace containing chosen options
        :return:
        """
        options = vars(nsp)
        for key in options:
            if options.get(key):
                self.intervention = key
                self.wf_request_id = options.get(key)[0]

    def intervene_fail(self, wf_id: int):
        """
        Send the workflow-fail message.

        :param wf_id: workflow ID
        :return:
        """
        logger.info("SENDING WORKFLOW FAIL MESSAGE!")
        wf_req = self.info.lookup_workflow_request(wf_id)
        failed_msg = {
            "service": "workflow",
            "routing_key": "workflow",
            "subject": wf_req,
            "type": "workflow-failed",
        }
        self.message_router.send_message(**failed_msg)

    def resolve_fail(self, wf_id: int):
        """
        Report whether this workflow request has failed as expected.

        :param wf_id: workflow ID
        :return:
        """
        wf_req = self.info.lookup_workflow_request(wf_id)
        check_wf_status = wf_req.state == "Failed"
        execution = self.cinfo.lookup_execution_by_workflow_request_id(wf_id)
        check_ex_status = execution.state == "Failed"
        check_req_status = execution.capability_request.state == "Failed"

        if all([check_wf_status, check_ex_status, check_req_status]):
            logger.info(
                f"Workflow request #{wf_id} and its associated capability execution, "
                f"#{execution.id},"
                f" and request, #{execution.capability_request.id}, have been successfully failed."
            )
        else:
            logger.info(
                f"Cannot confirm state change for workflow #{wf_id} "
                f"and its associated execution and request."
                f"Please check the database."
            )

    def check_status(self, wf_id: int) -> bool:
        """
        Report whether the request failed as expected.

        :param wf_id: workflow ID
        :return:
        """
        wf_req = self.info.lookup_workflow_request(wf_id)
        if wf_req is None:
            logger.info(f"No Workflow #{wf_id} found!")
        else:
            check_wf_status = wf_req.state == "Failed"
            execution = self.cinfo.lookup_execution_by_workflow_request_id(wf_id)
            check_ex_status = execution.state == "Failed"
            check_req_status = execution.capability_request.state == "Failed"

            if all([check_wf_status, check_ex_status, check_req_status]):
                logger.info(f"Workflow #{wf_id} and its associated execution and request " f"are already Failed.")
                return False
            if check_wf_status is True and (False in [check_ex_status, check_req_status]):
                logger.info(f"Workflow #{wf_id} is already Failed, but its execution " f"and/or request are not. ")
                choice = input("Do you wish to proceed? Type 'y' to proceed or 'n' to cancel. ")
                if choice == "y":
                    return True
                return False
            return True

    def intercede(self):
        """
        Intercede in the workflow request.

        :return:
        """
        if self.intervention == "force_fail":
            logger.info(f"Intervening with workflow request #{self.wf_request_id}....")
            if self.check_status(self.wf_request_id):
                self.intervene_fail(self.wf_request_id)
                self.resolve_fail(self.wf_request_id)
        else:
            logger.info("Unknown argument")


def parser() -> argparse.ArgumentParser:
    """
    Initialize argument parser for CLI

    :return: Argument Parser
    """
    arg_parser = argparse.ArgumentParser(
        description="System Mediator allows intervention in currently processing workspaces jobs",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    arg_parser.add_argument(
        "--force-fail",
        nargs=1,
        action="store",
        required=False,
        help="Force a specified workflow request into the failed state. " "Requires a workflow request id.",
    )
    return arg_parser


def make_session() -> Session:
    """
    Connect to Workspaces database

    :return: session
    """
    settings = CapoConfig().settings("metadataDatabase")
    url = settings.jdbcUrl.replace("jdbc:", "").replace("://", f"://{settings.jdbcUsername}:{settings.jdbcPassword}@")
    try:
        engine = create_engine(url)
    except argparse.ArgumentError as err:
        logger.error(f"ERROR: Invalid database URL {url}.\n{err}")
        sys.exit(1)

    session = sessionmaker(bind=engine)
    return session()


def main():
    """
    Launch the workspaces system mediator.

    :return:
    """
    logger.info("**************************************")
    logger.info("* WELCOME TO THE WS SYSTEM MEDIATOR! *")
    logger.info("**************************************")

    args = parser().parse_args()
    session = make_session()

    Arbiter(WorkflowInfo(session), CapabilityInfo(session), args).intercede()
    session.close()
