"""
Tests that test the view functionality of our capability REST API.
The logic can be found in `capability/views/capability.py`.
"""
import http

from pyramid.config import Configurator
from pyramid.httpexceptions import (
    HTTPBadRequest,
    HTTPExpectationFailed,
    HTTPNotFound,
    HTTPPreconditionFailed,
)
from pyramid.testing import DummyRequest


def test_view_capability(test_config: Configurator, request_null_capability: DummyRequest):
    """
    Tests the view_capability view to make sure it properly returns the info of a capability

    :param test_config: Dummy Pyramid Configurator object set up for testing
    :param request_null_capability: Dummy Pyramid request object set up with mocked DB access
    supporting the null capability
    """
    from capability.views.capability import view_capability

    expected_response = {
        "type": "Capability",
        "enabled": True,
        "max_jobs": 2,
        "name": "null",
        "steps": "test",
        "has_image_products": False,
        "paused": False
    }
    request_null_capability.matchdict["capability_name"] = "null"
    response = view_capability(request_null_capability)
    assert response.status_code == http.HTTPStatus.OK
    assert response.json_body == expected_response


def test_view_capability_not_found(
    test_config: Configurator, request_null_capability: DummyRequest
):
    """
    Tests the view capability view to make sure it properly returns a 404 Not Found exception
    when a capability does not exist

    :param test_config: Dummy Pyramid Configurator object set up for testing
    :param request_null_capability: Dummy Pyramid request object set up with mocked DB access
    supporting the null capability
    """
    from capability.views.capability import view_capability

    request_null_capability.matchdict["capability_name"] = "does_not_exist"
    response = view_capability(request_null_capability)
    assert response.status_code == http.HTTPStatus.NOT_FOUND
    assert type(response) is HTTPNotFound


def test_create_capability(test_config: Configurator, request_null_capability: DummyRequest):
    """
    Tests the create capability view to make sure it properly supports creation of capabilities

    :param test_config: Dummy Pyramid Configurator object set up for testing
    :param request_null_capability: Dummy Pyramid request object set up with mocked DB access
    supporting the null capability
    """
    from capability.views.capability import create_capability

    request_null_capability.json_body = {
        "capability_name": "test_create",
        "steps": "test",
        "max_jobs": 1,
        "has_image_products": False
    }

    # Assert test capability not in list of capabilities (mocked)
    assert not request_null_capability.capability_info.lookup_capability("test")   

    response = create_capability(request_null_capability)
    assert response.status_code == http.HTTPStatus.OK
    expected_response = {
        "type": "Capability",
        "enabled": True,
        "max_jobs": 1,
        "name": "test_create",
        "steps": "test",
        "has_image_products": False,
        "paused": False
    }
    assert response.json_body == expected_response
    # Assert test capability has been added to list of capabilities (mocked)
    assert request_null_capability.capability_info.lookup_capability("test_create")


def test_create_capability_error(test_config: Configurator, request_null_capability: DummyRequest):
    """
    Tests that the create_capability view correctly responds with exceptions given bad input

    :param test_config: Dummy Pyramid Configurator object set up for testing
    :param request_null_capability: Dummy Pyramid request object set up with mocked DB access
    supporting the null capability
    """
    from capability.views.capability import create_capability

    request_null_capability.json_body = {"capability_name": "test"}
    response_bad_params = create_capability(request_null_capability)
    assert response_bad_params.status_code == http.HTTPStatus.BAD_REQUEST
    assert type(response_bad_params) is HTTPBadRequest

    request_null_capability.json_body = {"capability_name": "null", "steps": "null", "max_jobs": 69, "has_image_products": False}
    response_already_exists = create_capability(request_null_capability)
    assert response_already_exists.status_code == http.HTTPStatus.PRECONDITION_FAILED
    assert type(response_already_exists) is HTTPPreconditionFailed


def test_edit_capability(test_config: Configurator, request_null_capability: DummyRequest):
    """
    Tests that the edit_capability view correctly supports editing of capabilities

    :param test_config: Dummy Pyramid Configurator object set up for testing
    :param request_null_capability: Dummy Pyramid request object set up with mocked DB access
    supporting the null capability
    """
    from capability.views.capability import edit_capability

    request_null_capability.matchdict["capability_name"] = "null"
    request_null_capability.json_body = {"steps": "test", "max_jobs": 3, "has_image_products": False}
    response_all_params = edit_capability(request_null_capability)
    assert response_all_params.status_code == http.HTTPStatus.OK

    request_null_capability.json_body = {"max_jobs": 3}
    response_jobs_only = edit_capability(request_null_capability)
    assert response_jobs_only.status_code == http.HTTPStatus.OK

    request_null_capability.json_body = {"steps": "steps steps steps"}
    response_steps_only = edit_capability(request_null_capability)
    assert response_steps_only.status_code == http.HTTPStatus.OK


def test_edit_capability_error(test_config: Configurator, request_null_capability: DummyRequest):
    """
    Tests that the edit_capability view correctly responds with exceptions

    :param test_config: Dummy Pyramid Configurator object set up for testing
    :param request_null_capability: Dummy Pyramid request object set up with mocked DB access
    supporting the null capability
    """
    from capability.views.capability import edit_capability

    request_null_capability.matchdict["capability_name"] = "not_found"
    request_null_capability.json_body = {"steps": "test", "max_jobs": 3}
    response_cap_not_exists = edit_capability(request_null_capability)
    assert response_cap_not_exists.status_code == http.HTTPStatus.PRECONDITION_FAILED
    assert type(response_cap_not_exists) is HTTPPreconditionFailed

    request_null_capability.matchdict["capability_name"] = "null"
    request_null_capability.json_body = {}
    response_empty = edit_capability(request_null_capability)
    assert response_empty.status_code == http.HTTPStatus.BAD_REQUEST
    assert type(response_empty) is HTTPBadRequest

    request_null_capability.matchdict["capability_name"] = "error"
    request_null_capability.json_body = {"steps": "test", "max_jobs": 3}
    response_edit_failed = edit_capability(request_null_capability)
    assert response_edit_failed.status_code == http.HTTPStatus.EXPECTATION_FAILED
    assert type(response_edit_failed) is HTTPExpectationFailed


def test_enable_capability(test_config: Configurator, request_null_capability: DummyRequest):
    """
    Tests that capabilities can be properly enabled

    :param test_config: Dummy Pyramid Configurator object set up for testing
    :param request_null_capability: Dummy Pyramid request object set up with mocked DB access
    supporting the null capability
    """
    from capability.views.capability import enable_capability

    request_null_capability.matchdict["capability_name"] = "null"
    response = enable_capability(request_null_capability)
    assert response.status_code == http.HTTPStatus.OK


def test_enable_capability_error(test_config: Configurator, request_null_capability: DummyRequest):
    """
    Tests that enable_capability view properly responds with HTTP exceptions given bad input

    :param test_config: Dummy Pyramid Configurator object set up for testing
    :param request_null_capability: Dummy Pyramid request object set up with mocked DB access
    supporting the null capability
    """
    from capability.views.capability import enable_capability

    request_null_capability.matchdict["capability_name"] = "does_not_exist"
    response_does_not_exist = enable_capability(request_null_capability)
    assert response_does_not_exist.status_code == http.HTTPStatus.PRECONDITION_FAILED
    assert type(response_does_not_exist) is HTTPPreconditionFailed
    request_null_capability.matchdict["capability_name"] = "error"
    response_enable_failed = enable_capability(request_null_capability)
    assert response_enable_failed.status_code == http.HTTPStatus.EXPECTATION_FAILED
    assert type(response_enable_failed) is HTTPExpectationFailed


def test_disable_capability(test_config: Configurator, request_null_capability: DummyRequest):
    """
    Tests that capabilities can be properly disabled

    :param test_config: Dummy Pyramid Configurator object set up for testing
    :param request_null_capability: Dummy Pyramid request object set up with mocked DB access
    supporting the null capability
    """
    from capability.views.capability import disable_capability

    request_null_capability.matchdict["capability_name"] = "null"
    response = disable_capability(request_null_capability)
    assert response.status_code == http.HTTPStatus.OK


def test_disable_capability_error(test_config: Configurator, request_null_capability: DummyRequest):
    """
    Tests that disable_capability view properly responds with HTTP exceptions given bad input

    :param test_config: Dummy Pyramid Configurator object set up for testing
    :param request_null_capability: Dummy Pyramid request object set up with mocked DB access
    supporting the null capability
    """
    from capability.views.capability import disable_capability

    request_null_capability.matchdict["capability_name"] = "does_not_exist"
    response_does_not_exist = disable_capability(request_null_capability)
    assert response_does_not_exist.status_code == http.HTTPStatus.PRECONDITION_FAILED
    assert type(response_does_not_exist) is HTTPPreconditionFailed
    request_null_capability.matchdict["capability_name"] = "error"
    response_disable_failed = disable_capability(request_null_capability)
    assert response_disable_failed.status_code == http.HTTPStatus.EXPECTATION_FAILED
    assert type(response_disable_failed) is HTTPExpectationFailed
