import abc
import json
import logging
from typing import Dict

import requests
from aat_wrest.observation_wrester import WrestObservationMetadata

from aat_wrest.utilities import MDDBConnector
from messaging.router import Router, on_message
from pycapo import CapoConfig
from requests import Response

from workspaces.system.services.interfaces import ArchiveServiceIF

logger = logging.getLogger(__name__)

"""
Standard calibrations are generated as ingestions complete. There is a Capo property
that controls whether the calibration requests are generated, generated and submitted, 
or just ignored. This is based on the earlier CIPL workflow in the archive. During 
array configuration changes, DAs suspend the executions but allow requests to be made,
because move configurations are not well-behaved with the standard recipes, so they 
require some manual intervention. Additionally, during various failure modes, the DAs
will completely ignore messages coming from ingest.

Since the Workspaces system will be deployed in parallel with the legacy CIPL 
workflow, we need a lever to control which one is handling standard calibrations
(or for testing, perhaps both). For this reason we may want to completely ignore 
ingestion-complete messages in one system or the other.

CIPL already implemented this functionality with a simple state in Capo. The state can
either be RUN, PAUSE or STOP. RUN (in essence) both creates and submits workflows,
whereas PAUSE creates but does not submit workflows, and STOP ignores the ingestion-
complete event entirely. So we duplicate this functionality here under a different key
so that both systems are configured the same way, to simplify life on the Capo editor.

What follows is an example of the Gang of Four Design Pattern "State." This is just
the state of the standard calibration listener, not to be confused with capability
or capability request states. The RunStates in this module shouldn't ever escape;
they're just an implementation detail in this module. By using this pattern we can
simplify the handling of the event, since we simply look up the state we're in and 
ask it to deal with the event somehow.
"""


class StandardCalibrationRunState(abc.ABC):
    """
    Represents a run state that the archive service uses to determine
    what to do in response to an ingest-complete message arriving.
    """

    @abc.abstractmethod
    def ingest_complete(self, locator: str, metadata: Dict[str, str]):
        """
        Handle an ingestion-complete message from the archive
        :param locator:  the science product locator of this data
        :param metadata: some metadata about this request
        :return:
        """
        pass


class RunState(StandardCalibrationRunState):
    """
    The RUN state: create the request, and also start it executing.
    """

    def __init__(self, manager: "StandardCalibrationManager"):
        self.manager = manager

    def ingest_complete(self, spl: str, metadata: Dict[str, str]):
        logger.debug(f"(RUN) Creating and submitting standard calibration request for {spl}")
        request_id = self.manager.create_stdcal_request(spl, metadata)
        self.manager.submit_stdcal_request(request_id)
        logger.info(f"(RUN) Submitted standard calibration request #{request_id} for {spl}")


class PauseState(StandardCalibrationRunState):
    """
    The PAUSE state: create the request, but don't immediately start it
    """

    def __init__(self, manager: "StandardCalibrationManager"):
        self.manager = manager

    def ingest_complete(self, spl: str, metadata: Dict[str, str]):
        logger.debug(f"(PAUSE) Creating standard calibration request for {spl}")
        request_id = self.manager.create_stdcal_request(spl, metadata)
        logger.info(f"(PAUSE) Created standard calibration request #{request_id} for {spl}")


class StopState(StandardCalibrationRunState):
    """
    The STOP state: drop the ingestion-complete message on the floor and do nothing with it.
    """

    def __init__(self, manager: "StandardCalibrationManager"):
        self.manager = manager

    def ingest_complete(self, spl: str, metadata: Dict[str, str]):
        # we're not doing anything in response to this event here
        logger.info(f"(STOP) Ignoring ingestion-complete message for {spl}")


class StandardCalibrationManager:
    def __init__(self):
        self.STATES = {"RUN": RunState(self), "STOP": StopState(self), "PAUSE": PauseState(self)}

    def ingestion_complete(self, spl: str, metadata: Dict[str, str]):
        self.state.ingest_complete(spl, metadata)

    @property
    def state(self) -> StandardCalibrationRunState:
        """
        Return the standard calibration state we're currently in
        :return:
        """
        return self.STATES[self.state_name]

    @property
    def state_name(self) -> str:
        return CapoConfig().settings("edu.nrao.workspaces.StandardCalibrationSettings").runState

    @staticmethod
    def create_stdcal_request(spl: str, metadata: Dict[str, str]) -> int:
        """
        Creates a new standard calibration with information given from an ingestion-completed event

        :param spl: Science product locator of observation
        :param metadata: Dict with the following information inside:
            - sdm_id: SDM ID (sometimes called fileset id)
            - bands: Radio bands for observation
            - array_config: Configuration of array telescopes
            - obs_start_time: Start time of observation
            - obs_end_time: End time of observation
        :return: HTTP Response for attempt to create std_calibration request
        """
        capability_url = CapoConfig().settings("edu.nrao.workspaces.CapabilitySettings").serviceUrl
        analyst_email = CapoConfig().settings("edu.nrao.workspaces.NotificationSettings").analystEmail
        parameters = {
            "parameters": {
                "product_locator": spl,
                "user_email": analyst_email,
                "metadata": metadata,
            }
        }

        result = requests.post(f"{capability_url}/capability/std_calibration/request/create", json=parameters)
        result.raise_for_status()
        return result.json()["id"]

    @staticmethod
    def submit_stdcal_request(request_id: int):
        capability_url = CapoConfig().settings("edu.nrao.workspaces.CapabilitySettings").serviceUrl
        result = requests.post(f"{capability_url}/capability/request/{request_id}/submit")
        result.raise_for_status()


class ArchiveService(ArchiveServiceIF):
    """
    The archive service: clients access this to interface with the archive services
    """

    def __init__(self):
        self.stdcal = StandardCalibrationManager()
        self.router = Router("archive")
        self.router.register(self)

    @on_message(message="Ingestion complete")
    def on_ingestion_complete(self, **message: Dict):
        """
        Accept a ingestion-complete message from the archive

        :param message: the ingestion-complete message
        """

        log_data = message["logData"]

        # log message for Demo
        logger.info(f"ingestion-complete event: {log_data}")

        execblock_id = log_data["execblock_id"]
        sdm_id = log_data["fileset_id"]
        project_code = log_data["project_code"]

        wrester = WrestObservationMetadata(MDDBConnector(), sdm_id=sdm_id)
        metadata = json.loads(wrester.wrest_observation_info())
        metadata["sdm_id"] = sdm_id
        spl = metadata["spl"]
        del metadata["spl"]

        self.stdcal.ingestion_complete(spl, metadata)

        return spl, execblock_id, sdm_id, project_code
